@@ -25,14 +25,15 @@
 
 #define G_LOG_DOMAIN "CallsManager"
 
-#include "config.h"
+#include "calls-config.h"
 
-#include "calls-application.h"
-#include "calls-account-provider.h"
+#include "calls-account.h"
 #include "calls-contacts-provider.h"
 #include "calls-manager.h"
+#include "calls-message-source.h"
 #include "calls-provider.h"
 #include "calls-settings.h"
+#include "calls-ui-call-data.h"
 #include "calls-ussd.h"
 
 #include "enum-types.h"
@@ -40,177 +41,262 @@
 #include <glib/gi18n.h>
 #include <libpeas/peas.h>
 
-struct _CallsManager
-{
-  GObject parent_instance;
+/**
+ * SECTION:manager
+ * @short_description: Central management object
+ * @Title: CallsManager
+ *
+ * #CallsManager is a singleton that manages lists of loaded #CallsProvider,
+ * #CallsOrigin and #CallsCall objects. It keeps track of which #CallsOrigin
+ * supports which protocol. It also checks which #CallsCall are ongoing and
+ * emits signals for the UI and other parts of the application to act on.
+ */
 
-  GHashTable *providers;
-  /* This is the protocols supported in principle. This is collected from the loaded
-     providers and does not imply that there are any origins able to handle a given protocol.
-     See origins_by_protocol for a GListStore of suitable origins per protocol.
-  */
-  GPtrArray *supported_protocols;
+static const char * const protocols[] = {
+  "tel",
+  "sip",
+  "sips"
+};
 
-  GListStore *origins;
+struct _CallsManager {
+  GObject                parent_instance;
+
+  GHashTable            *providers;
+
+  GListStore            *origins;
   /* origins_by_protocol maps protocol names to GListStore's of suitable origins */
-  GHashTable *origins_by_protocol;
+  GHashTable            *origins_by_protocol;
+  /* dial_actions_by_protocol maps protocol names to GSimpleActions */
+  GHashTable            *dial_actions_by_protocol;
+
+  /* map CallsCall to CallsUiCallData */
+  GHashTable            *calls;
 
   CallsContactsProvider *contacts_provider;
 
-  CallsManagerState state;
-  CallsSettings *settings;
+  CallsManagerFlags      state_flags;
+  CallsSettings         *settings;
 };
 
-G_DEFINE_TYPE (CallsManager, calls_manager, G_TYPE_OBJECT);
+static void
+calls_manager_message_source_interface_init (CallsMessageSourceInterface *iface)
+{
+}
+
+G_DEFINE_TYPE_WITH_CODE (CallsManager, calls_manager, G_TYPE_OBJECT,
+                         G_IMPLEMENT_INTERFACE (CALLS_TYPE_MESSAGE_SOURCE,
+                                                calls_manager_message_source_interface_init))
 
 enum {
   PROP_0,
-  PROP_STATE,
+  PROP_STATE_FLAGS,
   PROP_LAST_PROP,
 };
 static GParamSpec *props[PROP_LAST_PROP];
 
 
 enum {
-  SIGNAL_CALL_ADD,
-  SIGNAL_CALL_REMOVE,
-  /* TODO: currently this event isn't emitted since the plugins don't give use
-   * a usable error or error message. */
-  SIGNAL_ERROR,
+  UI_CALL_ADDDED,
+  UI_CALL_REMOVED,
   USSD_ADDED,
   USSD_CANCELLED,
   USSD_STATE_CHANGED,
   PROVIDERS_CHANGED,
   SIGNAL_LAST_SIGNAL,
 };
-static guint signals [SIGNAL_LAST_SIGNAL];
+static guint signals[SIGNAL_LAST_SIGNAL];
+
 
 static void
-set_state (CallsManager *self, CallsManagerState state)
+set_state_flags (CallsManager *self, CallsManagerFlags state_flags)
 {
-  if (self->state == state)
+  if (self->state_flags == state_flags)
     return;
 
-  self->state = state;
+  self->state_flags = state_flags;
 
-  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_STATE]);
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_STATE_FLAGS]);
 }
 
+
 static void
-update_state (CallsManager *self)
+update_state_flags (CallsManager *self)
 {
-  guint n_items;
   GHashTableIter iter;
-  gpointer key;
   gpointer value;
+  CallsManagerFlags state_flags = CALLS_MANAGER_FLAGS_UNKNOWN;
 
   g_assert (CALLS_IS_MANAGER (self));
 
-  if (g_hash_table_size (self->providers) == 0) {
-    set_state (self, CALLS_MANAGER_STATE_NO_PROVIDER);
+  g_hash_table_iter_init (&iter, self->providers);
+
+  while (g_hash_table_iter_next (&iter, NULL, &value)) {
+    CallsProvider *provider = CALLS_PROVIDER (value);
+
+    if (calls_provider_is_modem (provider)) {
+      state_flags |= CALLS_MANAGER_FLAGS_HAS_CELLULAR_PROVIDER;
+
+      if (calls_provider_is_operational (provider))
+        state_flags |= CALLS_MANAGER_FLAGS_HAS_CELLULAR_MODEM;
+    } else {
+      state_flags |= CALLS_MANAGER_FLAGS_HAS_VOIP_PROVIDER;
+
+      if (calls_provider_is_operational (provider))
+        state_flags |= CALLS_MANAGER_FLAGS_HAS_VOIP_ACCOUNT;
+    }
+  }
+  set_state_flags (self, state_flags);
+}
+
+
+static void
+on_dial_protocol_activated (GSimpleAction *action,
+                            GVariant      *parameter,
+                            CallsManager  *self)
+{
+  GApplication *application;
+  CallsOrigin *origin;
+  g_autofree char *target = NULL;
+  g_autofree char *origin_id = NULL;
+
+  g_variant_get (parameter, "(ss)", &target, &origin_id);
+  origin = calls_manager_get_origin_by_id (self, origin_id);
+
+  if (origin) {
+    calls_origin_dial (origin, target);
     return;
   }
 
-  g_hash_table_iter_init (&iter, self->providers);
+  if (origin_id && *origin_id)
+    g_debug ("Origin ID '%s' given, but was not found for call to '%s'",
+             origin_id, target);
 
-  while (g_hash_table_iter_next (&iter, &key, &value)) {
-    CallsProvider *provider = CALLS_PROVIDER (value);
-
-    if (calls_provider_is_modem (provider) && !calls_provider_is_operational (provider)) {
-      set_state (self, CALLS_MANAGER_STATE_NO_VOICE_MODEM);
-      return;
-    }
+  /* fall back to the default action if we could not determine origin to place call from */
+  application = g_application_get_default ();
+  if (!application) {
+    g_warning ("Could not get default application, cannot activate action '%s'",
+               g_action_get_name (G_ACTION (action)));
+    return;
   }
 
-  n_items = g_list_model_get_n_items (G_LIST_MODEL (self->origins));
-
-  if (n_items)
-    set_state (self, CALLS_MANAGER_STATE_READY);
-  else
-    set_state (self, CALLS_MANAGER_STATE_NO_ORIGIN);
+  g_action_group_activate_action (G_ACTION_GROUP (application),
+                                  "dial",
+                                  g_variant_new_string (target));
 }
 
-static gboolean
-check_supported_protocol (CallsManager *self,
-                          const char   *protocol)
-{
-  guint index;
-  g_assert (CALLS_IS_MANAGER (self));
-  g_assert (protocol);
 
-  if (self->supported_protocols->len > 0)
-    return g_ptr_array_find_with_equal_func (self->supported_protocols,
-                                             protocol,
-                                             g_str_equal,
-                                             &index);
-
-  return FALSE;
-}
-
-/* This function will update self->supported_protocols from available provider plugins */
 static void
-update_protocols (CallsManager *self)
+update_protocol_dial_actions (CallsManager *self)
 {
-  GHashTableIter iter;
-  gpointer key, value;
-  const char * const *protocols;
-
   g_assert (CALLS_IS_MANAGER (self));
 
-  g_ptr_array_remove_range (self->supported_protocols,
-                            0, self->supported_protocols->len);
+  for (guint i = 0; i < G_N_ELEMENTS (protocols); i++) {
+    GSimpleAction *action = g_hash_table_lookup (self->dial_actions_by_protocol,
+                                                 protocols[i]);
+    GListModel *protocol_origin = g_hash_table_lookup (self->origins_by_protocol,
+                                                       protocols[i]);
+    /* TODO take into account if origin is active: modem registered or VoIP account online */
+    gboolean action_enabled = g_list_model_get_n_items (protocol_origin) > 0;
 
-  g_hash_table_iter_init (&iter, self->providers);
-  while (g_hash_table_iter_next (&iter, &key, &value)) {
-    const char *name = key;
-    CallsProvider *provider = CALLS_PROVIDER (value);
+    g_simple_action_set_enabled (action, action_enabled);
+  }
+}
 
-    protocols = calls_provider_get_protocols (provider);
 
-    if (protocols == NULL) {
-      g_debug ("Plugin %s does not provide any protocols", name);
-      continue;
-    }
-    for (guint i = 0; protocols[i] != NULL; i++) {
-      if (!check_supported_protocol (self, protocols[i]))
-        g_ptr_array_add (self->supported_protocols, g_strdup (protocols[i]));
+static void
+on_message (CallsMessageSource *source,
+            const char         *message,
+            GtkMessageType      message_type,
+            CallsManager       *self)
+{
+  g_autofree char *notification = NULL;
 
-      if (!g_hash_table_contains (self->origins_by_protocol, protocols[i])) {
-        /* Add a new GListStore if there's none already.
-         * Actually adding origins to self->origins_by_protocol is done
-         * in rebuild_origins_by_protocol()
-         */
-        GListStore *store = g_list_store_new (CALLS_TYPE_ORIGIN);
-        g_hash_table_insert (self->origins_by_protocol,
-                             g_strdup (protocols[i]),
-                             store);
-      }
-    }
+  g_assert (CALLS_IS_MESSAGE_SOURCE (source));
+  g_assert (CALLS_IS_MANAGER (self));
+
+  /* Prefix the message with the name of the source, if known */
+  if (CALLS_IS_ACCOUNT (source)) {
+    notification = g_strdup_printf ("%s: %s",
+                                    calls_account_get_address (CALLS_ACCOUNT (source)),
+                                    message);
   }
 
-  update_state (self);
+  calls_message_source_emit_message (CALLS_MESSAGE_SOURCE (self),
+                                     notification ?: message,
+                                     message_type);
 }
 
+
 static void
 add_call (CallsManager *self, CallsCall *call, CallsOrigin *origin)
 {
+  CallsUiCallData *call_data;
+  g_autofree char *origin_id = NULL;
+
   g_return_if_fail (CALLS_IS_MANAGER (self));
   g_return_if_fail (CALLS_IS_ORIGIN (origin));
   g_return_if_fail (CALLS_IS_CALL (call));
 
-  g_signal_emit (self, signals[SIGNAL_CALL_ADD], 0, call, origin);
+  origin_id = calls_origin_get_id (origin);
+  call_data = calls_ui_call_data_new (call, origin_id);
+  g_hash_table_insert (self->calls, call, call_data);
+
+  g_signal_emit (self, signals[UI_CALL_ADDDED], 0, call_data);
 }
 
+
+struct CallsRemoveData {
+  CallsManager *manager;
+  CallsCall    *call;
+};
+
+
+static gboolean
+on_remove_delayed (struct CallsRemoveData *data)
+{
+  CallsUiCallData *call_data;
+
+  g_assert (CALLS_IS_MANAGER (data->manager));
+  g_assert (CALLS_IS_CALL (data->call));
+
+  call_data = g_hash_table_lookup (data->manager->calls, data->call);
+  if (!call_data) {
+    g_warning ("Could not find call %s in UI call hash table",
+               calls_call_get_id (data->call));
+  } else {
+    g_signal_emit (data->manager, signals[UI_CALL_REMOVED], 0, call_data);
+    g_hash_table_remove (data->manager->calls, data->call);
+  }
+
+  g_free (data);
+
+  return G_SOURCE_REMOVE;
+}
+
+
+#define DELAY_CALL_REMOVE_MS 3000
 static void
 remove_call (CallsManager *self, CallsCall *call, gchar *reason, CallsOrigin *origin)
 {
+  struct CallsRemoveData *data;
+
   g_return_if_fail (CALLS_IS_MANAGER (self));
   g_return_if_fail (CALLS_IS_ORIGIN (origin));
   g_return_if_fail (CALLS_IS_CALL (call));
 
-  /* We ignore the reason for now, because it doesn't give any usefull information */
-  g_signal_emit (self, signals[SIGNAL_CALL_REMOVE], 0, call, origin);
+  data = g_new (struct CallsRemoveData, 1);
+  data->manager = self;
+  data->call = call;
+
+  /** Having the delay centrally managed makes sure our UI
+   *  and the DBus side stays consistent
+   */
+  g_timeout_add (DELAY_CALL_REMOVE_MS,
+                 G_SOURCE_FUNC (on_remove_delayed),
+                 data);
 }
+#undef DELAY_CALL_REMOVE_MS
+
 
 static void
 ussd_added_cb (CallsManager *self,
@@ -223,6 +309,7 @@ ussd_added_cb (CallsManager *self,
   g_signal_emit (self, signals[USSD_ADDED], 0, ussd, response);
 }
 
+
 static void
 ussd_cancelled_cb (CallsManager *self,
                    CallsUssd    *ussd,
@@ -234,6 +321,7 @@ ussd_cancelled_cb (CallsManager *self,
   g_signal_emit (self, signals[USSD_CANCELLED], 0, ussd);
 }
 
+
 static void
 ussd_state_changed_cb (CallsManager *self,
                        CallsUssd    *ussd)
@@ -244,6 +332,7 @@ ussd_state_changed_cb (CallsManager *self,
   g_signal_emit (self, signals[USSD_STATE_CHANGED], 0, ussd);
 }
 
+
 static void
 update_country_code_cb (CallsOrigin  *origin,
                         GParamSpec   *pspec,
@@ -261,10 +350,12 @@ update_country_code_cb (CallsOrigin  *origin,
   calls_settings_set_country_code (self->settings, country_code);
 }
 
+
 static void
 add_origin (CallsManager *self, CallsOrigin *origin)
 {
   g_autofree const char *name = NULL;
+
   g_assert (CALLS_IS_MANAGER (self));
   g_assert (CALLS_IS_ORIGIN (origin));
 
@@ -273,6 +364,11 @@ add_origin (CallsManager *self, CallsOrigin *origin)
 
   g_list_store_append (self->origins, origin);
 
+  g_signal_connect (origin,
+                    "message",
+                    G_CALLBACK (on_message),
+                    self);
+
   g_signal_connect_object (origin,
                            "notify::country-code",
                            G_CALLBACK (update_country_code_cb),
@@ -281,22 +377,23 @@ add_origin (CallsManager *self, CallsOrigin *origin)
   g_signal_connect_swapped (origin, "call-added", G_CALLBACK (add_call), self);
   g_signal_connect_swapped (origin, "call-removed", G_CALLBACK (remove_call), self);
 
-  if (CALLS_IS_USSD (origin))
-    {
-      g_signal_connect_swapped (origin, "ussd-added", G_CALLBACK (ussd_added_cb), self);
-      g_signal_connect_swapped (origin, "ussd-cancelled", G_CALLBACK (ussd_cancelled_cb), self);
-      g_signal_connect_swapped (origin, "ussd-state-changed", G_CALLBACK (ussd_state_changed_cb), self);
-    }
+  if (CALLS_IS_USSD (origin)) {
+    g_signal_connect_swapped (origin, "ussd-added", G_CALLBACK (ussd_added_cb), self);
+    g_signal_connect_swapped (origin, "ussd-cancelled", G_CALLBACK (ussd_cancelled_cb), self);
+    g_signal_connect_swapped (origin, "ussd-state-changed", G_CALLBACK (ussd_state_changed_cb), self);
+  }
 
   calls_origin_foreach_call (origin, (CallsOriginForeachCallFunc) add_call, self);
 }
 
+
 static void
 remove_call_cb (gpointer self, CallsCall *call, CallsOrigin *origin)
 {
   remove_call (self, call, NULL, origin);
 }
 
+
 static void
 remove_origin (CallsManager *self, CallsOrigin *origin)
 {
@@ -318,8 +415,6 @@ remove_origin (CallsManager *self, CallsOrigin *origin)
                origin);
   else
     g_list_store_remove (self->origins, position);
-
-  update_state (self);
 }
 
 /* rebuild_origins_by_protocols() when any origins were added or removed */
@@ -327,7 +422,7 @@ static void
 rebuild_origins_by_protocols (CallsManager *self)
 {
   GHashTableIter iter;
-  gpointer key, value;
+  gpointer value;
   guint n_origins;
 
   g_assert (CALLS_IS_MANAGER (self));
@@ -335,7 +430,7 @@ rebuild_origins_by_protocols (CallsManager *self)
   /* Remove everything */
   g_hash_table_iter_init (&iter, self->origins_by_protocol);
 
-  while (g_hash_table_iter_next (&iter, &key, &value)) {
+  while (g_hash_table_iter_next (&iter, NULL, &value)) {
     GListStore *store = G_LIST_STORE (value);
     g_list_store_remove_all (store);
   }
@@ -347,26 +442,29 @@ rebuild_origins_by_protocols (CallsManager *self)
     g_autoptr (CallsOrigin) origin =
       g_list_model_get_item (G_LIST_MODEL (self->origins), i);
 
-    for (guint j = 0; j < self->supported_protocols->len; j++) {
-      char *protocol = g_ptr_array_index (self->supported_protocols, j);
+    for (guint j = 0; j < G_N_ELEMENTS (protocols); j++) {
       GListStore *store =
-        G_LIST_STORE (g_hash_table_lookup (self->origins_by_protocol, protocol));
+        G_LIST_STORE (g_hash_table_lookup (self->origins_by_protocol, protocols[j]));
 
       g_assert (store);
 
-      if (calls_origin_supports_protocol (origin, protocol))
+      if (calls_origin_supports_protocol (origin, protocols[j]))
         g_list_store_append (store, origin);
     }
   }
+
+  update_protocol_dial_actions (self);
 }
 
+
 static void
 remove_provider (CallsManager *self,
                  const char   *name)
 {
+  g_autoptr (CallsProvider) provider = NULL;
+
   GListModel *origins;
   guint n_items;
-  g_autoptr (CallsProvider) provider = NULL;
 
   g_assert (CALLS_IS_MANAGER (self));
   g_assert (name);
@@ -388,7 +486,7 @@ remove_provider (CallsManager *self,
   n_items = g_list_model_get_n_items (origins);
 
   for (guint i = 0; i < n_items; i++) {
-    g_autoptr(CallsOrigin) origin = NULL;
+    g_autoptr (CallsOrigin) origin = NULL;
 
     origin = g_list_model_get_item (origins, i);
     remove_origin (self, origin);
@@ -397,30 +495,30 @@ remove_provider (CallsManager *self,
   g_hash_table_remove (self->providers, name);
   calls_provider_unload_plugin (name);
 
-  update_protocols (self);
-  update_state (self);
   rebuild_origins_by_protocols (self);
+  update_state_flags (self);
 
   g_signal_emit (self, signals[PROVIDERS_CHANGED], 0);
 }
 
+
 static gboolean
 origin_found_in_any_provider (CallsManager *self,
                               CallsOrigin  *origin)
 {
   GHashTableIter iter;
-  gpointer key, value;
+  gpointer value;
 
   g_return_val_if_fail (CALLS_IS_MANAGER (self), FALSE);
   g_return_val_if_fail (CALLS_IS_ORIGIN (origin), FALSE);
 
   g_hash_table_iter_init (&iter, self->providers);
-  while (g_hash_table_iter_next (&iter, &key, &value)) {
+  while (g_hash_table_iter_next (&iter, NULL, &value)) {
     guint position;
     CallsProvider *provider = CALLS_PROVIDER (value);
     GListModel *origins = calls_provider_get_origins (provider);
 
-    if (origins && calls_find_in_store (origins,
+    if (origins && calls_find_in_model (origins,
                                         origin,
                                         &position))
       return TRUE;
@@ -430,7 +528,6 @@ origin_found_in_any_provider (CallsManager *self,
 }
 
 
-
 static void
 origin_items_changed_cb (GListModel   *model,
                          guint         position,
@@ -469,7 +566,7 @@ origin_items_changed_cb (GListModel   *model,
     g_warning ("Managed origins are not in sync anymore!");
   }
 
- skip_remove:
+skip_remove:
   for (i = 0; i < added; i++) {
     g_debug ("before adding: %d",
              g_list_model_get_n_items (G_LIST_MODEL (self->origins)));
@@ -483,9 +580,10 @@ origin_items_changed_cb (GListModel   *model,
   }
 
   rebuild_origins_by_protocols (self);
-  update_state (self);
+  update_state_flags (self);
 }
 
+
 static void
 add_provider (CallsManager *self, const gchar *name)
 {
@@ -507,8 +605,6 @@ add_provider (CallsManager *self, const gchar *name)
 
   g_hash_table_insert (self->providers, g_strdup (name), provider);
 
-  update_protocols (self);
-
   origins = calls_provider_get_origins (provider);
 
   g_signal_connect_object (origins, "items-changed",
@@ -521,6 +617,7 @@ add_provider (CallsManager *self, const gchar *name)
   g_signal_emit (self, signals[PROVIDERS_CHANGED], 0);
 }
 
+
 static void
 calls_manager_get_property (GObject    *object,
                             guint       property_id,
@@ -530,8 +627,8 @@ calls_manager_get_property (GObject    *object,
   CallsManager *self = CALLS_MANAGER (object);
 
   switch (property_id) {
-  case PROP_STATE:
-    g_value_set_enum (value, calls_manager_get_state (self));
+  case PROP_STATE_FLAGS:
+    g_value_set_flags (value, calls_manager_get_state_flags (self));
     break;
 
   default:
@@ -548,11 +645,10 @@ calls_manager_finalize (GObject *object)
 
   g_clear_object (&self->origins);
   g_clear_object (&self->contacts_provider);
-  g_clear_object (&self->settings);
 
   g_clear_pointer (&self->providers, g_hash_table_unref);
   g_clear_pointer (&self->origins_by_protocol, g_hash_table_unref);
-  g_clear_pointer (&self->supported_protocols, g_ptr_array_unref);
+  g_clear_pointer (&self->dial_actions_by_protocol, g_hash_table_unref);
 
   G_OBJECT_CLASS (calls_manager_parent_class)->finalize (object);
 }
@@ -566,37 +662,25 @@ calls_manager_class_init (CallsManagerClass *klass)
   object_class->get_property = calls_manager_get_property;
   object_class->finalize = calls_manager_finalize;
 
-  signals[SIGNAL_CALL_ADD] =
-   g_signal_new ("call-add",
-                 G_TYPE_FROM_CLASS (klass),
-                 G_SIGNAL_RUN_FIRST,
-                 0,
-                 NULL, NULL, NULL,
-                 G_TYPE_NONE,
-                 2,
-                 CALLS_TYPE_CALL,
-                 CALLS_TYPE_ORIGIN);
+  signals[UI_CALL_ADDDED] =
+    g_signal_new ("ui-call-added",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE,
+                  1,
+                  CALLS_TYPE_UI_CALL_DATA);
 
-  signals[SIGNAL_CALL_REMOVE] =
-   g_signal_new ("call-remove",
-                 G_TYPE_FROM_CLASS (klass),
-                 G_SIGNAL_RUN_FIRST,
-                 0,
-                 NULL, NULL, NULL,
-                 G_TYPE_NONE,
-                 2,
-                 CALLS_TYPE_CALL,
-                 CALLS_TYPE_ORIGIN);
-
-  signals[SIGNAL_ERROR] =
-   g_signal_new ("error",
-                 G_TYPE_FROM_CLASS (klass),
-                 G_SIGNAL_RUN_FIRST,
-                 0,
-                 NULL, NULL, NULL,
-                 G_TYPE_NONE,
-                 1,
-                 G_TYPE_STRING);
+  signals[UI_CALL_REMOVED] =
+    g_signal_new ("ui-call-removed",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE,
+                  1,
+                  CALLS_TYPE_UI_CALL_DATA);
 
   signals[USSD_ADDED] =
     g_signal_new ("ussd-added",
@@ -638,13 +722,13 @@ calls_manager_class_init (CallsManagerClass *klass)
                   G_TYPE_NONE,
                   0);
 
-  props[PROP_STATE] =
-    g_param_spec_enum ("state",
-                       "state",
-                       "The state of the Manager",
-                       CALLS_TYPE_MANAGER_STATE,
-                       CALLS_MANAGER_STATE_NO_PROVIDER,
-                       G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY);
+  props[PROP_STATE_FLAGS] =
+    g_param_spec_flags ("state",
+                        "state",
+                        "The state of the Manager",
+                        CALLS_TYPE_MANAGER_FLAGS,
+                        CALLS_MANAGER_FLAGS_UNKNOWN,
+                        G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY);
 
   g_object_class_install_properties (object_class, PROP_LAST_PROP, props);
 }
@@ -653,10 +737,13 @@ calls_manager_class_init (CallsManagerClass *klass)
 static void
 calls_manager_init (CallsManager *self)
 {
+  g_autoptr (GVariantType) variant_type = NULL;
+  GApplication *application;
   PeasEngine *peas;
   const gchar *dir;
+  g_autofree char *default_plugin_dir_provider = NULL;
 
-  self->state = CALLS_MANAGER_STATE_NO_PROVIDER;
+  self->state_flags = CALLS_MANAGER_FLAGS_UNKNOWN;
   self->providers = g_hash_table_new_full (g_str_hash,
                                            g_str_equal,
                                            g_free,
@@ -667,26 +754,71 @@ calls_manager_init (CallsManager *self)
                                                      g_free,
                                                      g_object_unref);
 
-  self->origins = g_list_store_new (calls_origin_get_type ());
-  self->supported_protocols = g_ptr_array_new_full (5, g_free);
+  for (guint i = 0; i < G_N_ELEMENTS (protocols); i++) {
+    GListStore *origin_store = g_list_store_new (calls_origin_get_type ());
+    g_hash_table_insert (self->origins_by_protocol,
+                         g_strdup (protocols[i]),
+                         origin_store);
+  }
 
-  self->settings = calls_settings_new ();
+  self->dial_actions_by_protocol = g_hash_table_new_full (g_str_hash,
+                                                          g_str_equal,
+                                                          g_free,
+                                                          g_object_unref);
+
+  application = g_application_get_default ();
+  variant_type = g_variant_type_new ("(ss)");
+
+  for (guint i = 0; i < G_N_ELEMENTS (protocols); i++) {
+    g_autofree char *action_name = g_strdup_printf ("dial-%s", protocols[i]);
+    GSimpleAction *action = g_simple_action_new (action_name, variant_type);
+    g_signal_connect (action,
+                      "activate",
+                      G_CALLBACK (on_dial_protocol_activated),
+                      self);
+
+    g_hash_table_insert (self->dial_actions_by_protocol,
+                         g_strdup (protocols[i]),
+                         g_object_ref (action));
+
+    /* Enable action if there are suitable origins */
+    g_simple_action_set_enabled (action, FALSE);
+
+    /* application can be NULL when running tests */
+    if (application)
+      g_action_map_add_action (G_ACTION_MAP (application), G_ACTION (action));
+  }
+
+  self->origins = g_list_store_new (calls_origin_get_type ());
+
+  /* This hash table only owns the value, not the key */
+  self->calls = g_hash_table_new_full (NULL, NULL, NULL, g_object_unref);
+
+  self->settings = calls_settings_get_default ();
   // Load the contacts provider
-  self->contacts_provider = calls_contacts_provider_new (self->settings);
+  self->contacts_provider = calls_contacts_provider_new ();
 
   peas = peas_engine_get_default ();
 
   dir = g_getenv ("CALLS_PLUGIN_DIR");
   if (dir && dir[0] != '\0') {
-    /** Add the directory to the search path. prepend_search_path() does not work
-     * as expected. see https://gitlab.gnome.org/GNOME/libpeas/-/issues/19
-     */
-    g_debug ("Adding %s to plugin search path", dir);
-    peas_engine_add_search_path (peas, dir, NULL);
+    g_autofree char *plugin_dir_provider = NULL;
+
+    plugin_dir_provider = g_build_filename (dir, "provider", NULL);
+
+    if (g_file_test (plugin_dir_provider, G_FILE_TEST_EXISTS)) {
+      g_debug ("Adding %s to plugin search path", plugin_dir_provider);
+      peas_engine_prepend_search_path (peas, plugin_dir_provider, NULL);
+    } else {
+      g_warning ("Not adding %s to plugin search path, because the directory doesn't exist. Check if env CALLS_PLUGIN_DIR is set correctly", plugin_dir_provider);
+    }
   }
 
-  peas_engine_add_search_path (peas, PLUGIN_LIBDIR, NULL);
-  g_debug ("Scanning for plugins in `%s'", PLUGIN_LIBDIR);
+  default_plugin_dir_provider = g_build_filename(PLUGIN_LIBDIR, "provider", NULL);
+  peas_engine_add_search_path (peas, default_plugin_dir_provider, NULL);
+  g_debug ("Scanning for plugins in `%s'", default_plugin_dir_provider);
+
+  peas_engine_rescan_plugins (peas);
 }
 
 
@@ -696,6 +828,7 @@ calls_manager_new (void)
   return g_object_new (CALLS_TYPE_MANAGER, NULL);
 }
 
+
 CallsManager *
 calls_manager_get_default (void)
 {
@@ -703,11 +836,12 @@ calls_manager_get_default (void)
 
   if (instance == NULL) {
     instance = calls_manager_new ();
-    g_object_add_weak_pointer (G_OBJECT (instance), (gpointer *)&instance);
+    g_object_add_weak_pointer (G_OBJECT (instance), (gpointer *) &instance);
   }
   return instance;
 }
 
+
 CallsContactsProvider *
 calls_manager_get_contacts_provider (CallsManager *self)
 {
@@ -716,6 +850,7 @@ calls_manager_get_contacts_provider (CallsManager *self)
   return self->contacts_provider;
 }
 
+
 void
 calls_manager_add_provider (CallsManager *self,
                             const char   *name)
@@ -726,6 +861,7 @@ calls_manager_add_provider (CallsManager *self,
   add_provider (self, name);
 }
 
+
 void
 calls_manager_remove_provider (CallsManager *self,
                                const char   *name)
@@ -734,9 +870,9 @@ calls_manager_remove_provider (CallsManager *self,
   g_return_if_fail (name);
 
   remove_provider (self, name);
-  update_protocols (self);
 }
 
+
 gboolean
 calls_manager_has_provider (CallsManager *self,
                             const char   *name)
@@ -747,6 +883,7 @@ calls_manager_has_provider (CallsManager *self,
   return !!g_hash_table_lookup (self->providers, name);
 }
 
+
 gboolean
 calls_manager_is_modem_provider (CallsManager *self,
                                  const char   *name)
@@ -763,14 +900,15 @@ calls_manager_is_modem_provider (CallsManager *self,
 }
 
 
-CallsManagerState
-calls_manager_get_state (CallsManager *self)
+CallsManagerFlags
+calls_manager_get_state_flags (CallsManager *self)
 {
-  g_return_val_if_fail (CALLS_IS_MANAGER (self), CALLS_MANAGER_STATE_UNKNOWN);
+  g_return_val_if_fail (CALLS_IS_MANAGER (self), CALLS_MANAGER_FLAGS_UNKNOWN);
 
-  return self->state;
+  return self->state_flags;
 }
 
+
 GListModel *
 calls_manager_get_origins (CallsManager *self)
 {
@@ -779,28 +917,21 @@ calls_manager_get_origins (CallsManager *self)
   return G_LIST_MODEL (self->origins);
 }
 
+
+/**
+ * calls_manager_get_calls:
+ * @self: A #CallsManager
+ *
+ * Returns: (transfer container): Returns a list of all known calls.
+ * The calls are objects of type #CallsUiCallData. Use g_list_free() when
+ * done using the list.
+ */
 GList *
 calls_manager_get_calls (CallsManager *self)
 {
-  GListModel *origins = NULL;
-  g_autoptr (GList) calls = NULL;
-  guint n_items = 0;
-
   g_return_val_if_fail (CALLS_IS_MANAGER (self), NULL);
 
-  origins = calls_manager_get_origins (self);
-  if (origins)
-    n_items = g_list_model_get_n_items (origins);
-
-  for (guint i = 0; i < n_items; i++)
-    {
-      g_autoptr (CallsOrigin) origin = NULL;
-
-      origin = g_list_model_get_item (origins, i);
-      calls = g_list_concat (calls, calls_origin_get_calls (origin));
-    }
-
-  return g_steal_pointer (&calls);
+  return g_hash_table_get_values (self->calls);
 }
 
 /**
@@ -812,51 +943,28 @@ calls_manager_get_calls (CallsManager *self)
 void
 calls_manager_hang_up_all_calls (CallsManager *self)
 {
-  g_autoptr (GList) calls = NULL;
+  GListModel *origins;
   GList *node;
   CallsCall *call;
+  uint n_items;
 
   g_return_if_fail (CALLS_IS_MANAGER (self));
 
-  calls = calls_manager_get_calls (self);
+  origins = G_LIST_MODEL (self->origins);
+  n_items = g_list_model_get_n_items (origins);
 
-  for (node = calls; node; node = node->next)
-    {
+  for (uint i = 0; i < n_items; i++) {
+    g_autoptr (CallsOrigin) origin = g_list_model_get_item (origins, i);
+    g_autoptr (GList) calls = calls_origin_get_calls (origin);
+
+    for (node = calls; node; node = node->next) {
       call = node->data;
       g_debug ("Hanging up on call %s", calls_call_get_name (call));
       calls_call_hang_up (call);
     }
+  }
 
-  g_debug ("Hanged up on all calls");
-}
-
-/**
- * calls_manager_has_active_call
- * @self: a #CallsManager
- *
- * Checks if @self has any active call
- *
- * Returns: %TRUE if there are active calls, %FALSE otherwise
- */
-gboolean
-calls_manager_has_active_call (CallsManager *self)
-{
-  g_autoptr (GList) calls = NULL;
-  GList *node;
-  CallsCall *call;
-
-  g_return_val_if_fail (CALLS_IS_MANAGER (self), FALSE);
-
-  calls = calls_manager_get_calls (self);
-
-  for (node = calls; node; node = node->next)
-    {
-      call = node->data;
-      if (calls_call_get_state (call) != CALLS_CALL_STATE_DISCONNECTED)
-        return TRUE;
-    }
-
-  return FALSE;
+  g_debug ("Hung up on all calls");
 }
 
 /**
@@ -864,7 +972,8 @@ calls_manager_has_active_call (CallsManager *self)
  * @self: The #CallsManager
  * @target: The target number/address
  *
- * Returns (transfer none): A #GListModel of suitable origins
+ * Returns (transfer none): A #GListModel of suitable origins as long
+ * as the protocol to be used for @target is supported, %NULL otherwise
  */
 GListModel *
 calls_manager_get_suitable_origins (CallsManager *self,
@@ -945,3 +1054,35 @@ calls_manager_get_settings (CallsManager *self)
 
   return self->settings;
 }
+
+/**
+ * calls_manager_get_origin_by_id:
+ * @self: The #CallsManager
+ * @origin_id: The id to use for the lookup
+ *
+ * Returns: (transfer none): The #CallsOrigin if found, %NULL otherwise
+ */
+CallsOrigin *
+calls_manager_get_origin_by_id (CallsManager *self,
+                                const char   *origin_id)
+{
+  uint n_origins;
+
+  g_return_val_if_fail (CALLS_IS_MANAGER (self), NULL);
+
+  /* TODO Turn this into a critical once https://gitlab.gnome.org/GNOME/calls/-/merge_requests/505 is in */
+  if (origin_id && *origin_id)
+    return NULL;
+
+  n_origins = g_list_model_get_n_items (G_LIST_MODEL (self->origins));
+  for (uint i = 0; i < n_origins; i++) {
+    g_autoptr (CallsOrigin) origin =
+      g_list_model_get_item (G_LIST_MODEL (self->origins), i);
+    g_autofree char *id = calls_origin_get_id (origin);
+
+    if (g_strcmp0 (id, origin_id) == 0)
+      return origin;
+  }
+
+  return NULL;
+}
